/*****************************************************************************
 *                                                                           *
 *  Copyright (c) 1993-1997 Elan Feingold (elan@jeeves.net)                  *
 *                                                                           *
 *     PERMISSION TO USE, COPY, MODIFY, AND TO DISTRIBUTE THIS SOFTWARE      *
 *     AND ITS DOCUMENTATION FOR ANY PURPOSE IS HEREBY GRANTED WITHOUT       *
 *     FEE, PROVIDED THAT THE ABOVE COPYRIGHT NOTICE APPEAR IN ALL           *
 *     COPIES AND MODIFIED COPIES AND THAT BOTH THAT COPYRIGHT NOTICE AND    *
 *     THIS PERMISSION NOTICE APPEAR IN SUPPORTING DOCUMENTATION.  THERE     *
 *     IS NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR      *
 *     ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS       *
 *     OR IMPLIED WARRANTY.                                                  *
 *                                                                           *
 *****************************************************************************/

#ifndef _AICLIENT
#define _AICLIENT

#include "types.h"

/* Private stuff */
#define DefineSpecies(callback, name, author, version, desc) \
void *(*__AI_Callback)(void *, Int32, void *) = callback; \
CString __strName = name; \
CString __strAuthor = author; \
CString __strVersion = version; \
CString __strDesc = desc;

/* Registering to observe the distributed object (for advanced clients
 * only -- actually this was only added because the client I was
 * porting over needed it, but it might be useful for others).
 * Preobservers get called before the object is changed, so that they
 * can see what changed.  Postobservers get called after the object is
 * changed.
 */

void AI_RegisterPreObserver(void (*PreCallback)(Int32, void *));
void AI_RegisterPostObserver(void (*PostCallback)(Int32, void *));

/* Commands sent to the callback */
#define AI_INIT_ONCE        0
#define AI_INIT_GAME        1
#define AI_INIT_TURN        2
#define AI_FORTIFY          3 /* (Number of armies) */
#define AI_PLACE            4
#define AI_ATTACK           5
#define AI_MOVE             6
#define AI_EXCHANGE_CARDS   7
#define AI_SERVER_MESSAGE   8 /* (Text of the message) */
#define AI_MESSAGE          9 /* (Text of the message) */
#define AI_MOVE_MANUAL     10 /* (Number of armies) */

/* Actions computer players can perform -- PUBLIC INTERFACE */
Flag AI_Place(Int32 iCountry, Int32 iNumArmies);
Flag AI_Attack(Int32 iSrcCountry, Int32 iDstCountry, 
	       Int32 iAttackMode, Int32 iDiceMode, Int32 iMoveMode);
Flag AI_Move(Int32 iSrcCountry, Int32 iDstCountry, Int32 iNumArmies);
Flag AI_ExchangeCards(Int32 *piCards);
Flag AI_SendMessage(Int32 iMessDest, CString strMessage);
Flag AI_EndTurn(void);

/* Attack modes */
#define ATTACK_ONCE      0
#define ATTACK_DOORDIE   1

/* Dice modes */
#define DICE_MAXIMUM     0
#define DICE_ONE         1
#define DICE_TWO         2
#define DICE_THREE       3

/* What to do with armies after taking a country.  If you wish to move
 * the maximum amount of armies into the country you might occupy with
 * the attack, pass ARMIES_MOVE_MAX.  If you want the minimum number
 * of armies moved, pass ARMIES_MOVE_MIN.  If you want finer control,
 * pass in ARMIES_MOVE_MANUAL, and your callback will be called with
 * the command AI_MOVE_MANUAL.
 */

#define ARMIES_MOVE_MAX    0
#define ARMIES_MOVE_MIN    1
#define ARMIES_MOVE_MANUAL 2

#endif
